import { assertEquals } from "https://deno.land/std@0.224.0/assert/mod.ts"
import { exists } from "https://deno.land/std@0.224.0/fs/mod.ts"

export type Todo = {
	tasks: Task[]
}

export type Location = {
	line: number
	column: number
}

export type Task = {
	title: string
	done: boolean
	tick_loc: Location
	subtasks: Task[]
	descriptions: string[]
}

function new_task(title: string, tick_loc: Location) {
	return {
		title,
		done: false,
		tick_loc,
		subtasks: [],
		descriptions: [],
	} as Task
}

export function parse_todo(content: string): Todo {
	const root = new_task("root", { line: -1, column: -1 })
	const stack = [[-1, root]] as [number, Task][]
	let line = 0
	for (const line_text of content.split("\n")) {
		const [indent, text] = unindent(line_text)
		if (text !== "") {
			while (stack[0][0] >= indent) stack.splice(0, 1)
			const parent = stack[0][1]
			const column = indent + 3
			if (text.startsWith("- [ ] ")) {
				const title = text.slice(6)
				const task = new_task(title, { line, column })
				parent.subtasks.push(task)
				stack.splice(0, 0, [indent, task])
			} else if (text.startsWith("- [")) {
				const title = text.slice(6)
				const task = new_task(title, { line, column })
				task.done = true
				parent.subtasks.push(task)
				stack.splice(0, 0, [indent, task])
			} else if (text.startsWith("- ")) {
				const description = text.slice(2)
				parent.descriptions.push(description)
			} else {
				const description = text
				parent.descriptions.push(description)
			}
		}
		line += 1
	}
	return { tasks: root.subtasks }
}

export async function read_todo(path: string) {
	// TODO create in repo and link to location
	if (!await exists(path)) await write_default_todo(path)
	const content = await Deno.readTextFile(path)
	const parsed = parse_todo(content)
	return parsed
}

async function write_default_todo(path: string) {
	const content = `
TODO
====

- [ ] Create tasks.

`
	await Deno.writeTextFile(path, content)
}

function unindent(line: string) {
	const indents = [" ", "\t"]
	let remainder = line
	let indent_count = 0
	while (true) {
		const indent = indents.find((i) => remainder.startsWith(i))
		if (indent === undefined) break
		remainder = remainder.substring(1)
		indent_count += 1
	}
	return [indent_count, remainder] as const
}

export function todo_completed(task: Task) {
	if (!task.done) return false
	for (const sub of task.subtasks) if (!todo_completed(sub)) return false
	return true
}

export function set_todo_tick(todo: Todo) {
	// TODO
}

Deno.test("test_parse_todo", () => {
	assertEquals(
		parse_todo(
			`

- [ ] arbre mort
- [o] lorem ipsum
    - dolor
    - [o] sed

- [o] B
- [ ] C

- [ ] D
    - arbre
- [ ] E
    - [o] F
        - [o] G
- machin

        `,
		),
		{
			tasks: [
				{
					title: "arbre mort",
					done: false,
					tick_loc: { line: 2, column: 3 },
					descriptions: [],
					subtasks: [],
				},
				{
					title: "lorem ipsum",
					done: true,
					tick_loc: { line: 3, column: 3 },
					descriptions: ["dolor"],
					subtasks: [
						{
							title: "sed",
							done: true,
							tick_loc: { line: 5, column: 7 },
							descriptions: [],
							subtasks: [],
						},
					],
				},
				{
					title: "B",
					done: true,
					tick_loc: { line: 7, column: 3 },
					descriptions: [],
					subtasks: [],
				},
				{
					title: "C",
					done: false,
					tick_loc: { line: 8, column: 3 },
					descriptions: [],
					subtasks: [],
				},
				{
					title: "D",
					done: false,
					tick_loc: { line: 10, column: 3 },
					descriptions: ["arbre"],
					subtasks: [],
				},
				{
					title: "E",
					done: false,
					tick_loc: { line: 12, column: 3 },
					descriptions: [],
					subtasks: [
						{
							title: "F",
							done: true,
							tick_loc: { line: 13, column: 7 },
							descriptions: [],
							subtasks: [
								{
									title: "G",
									done: true,
									tick_loc: { line: 14, column: 11 },
									descriptions: [],
									subtasks: [],
								},
							],
						},
					],
				},
			],
		},
	)
})
